import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.ImageIcon;

import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;

/**
 * This panel displays an image and periodically overlays a word balloon with
 * witty sayings.  The source of the balloon, the images, and the sayings are
 * all configurable.
 * @author Ted Faber &lt;faber@lunabase.org&gt;
 * @version 1.0
 */
public class Mojo extends JPanel {

    /**
     * This is a panel containing the word balloon that is laid over the Pic
     */
    class WordBalloon extends JPanel { 
	/** The funny saying, formatted for display */
	JTextArea q;
	/** The origin of the word balloon's tail */
	Point origin;

	/**
	 * Constructor.  Set the saying and the origin point.  The panel itself
	 * is sized to the applet size.
	 * @param s the saying to display (may be null)
	 * @param o the origin of the word balloon's tail (may be null)
	 */
	WordBalloon(String s, Point o) {
	    super(new BorderLayout());

	    Dimension base = Mojo.this.getSize();
	    int spacerX = Math.max((10 * base.width) / 400, 5);
	    int spacerY = Math.max((10 * base.height) / 400, 5);
	    Border eb = new EmptyBorder(spacerY, spacerX, spacerY, spacerX);
	    int cols =
		(8*base.width / getFontMetrics(getFont()).getMaxAdvance())/10;

	    origin = (o != null) ? (Point) o.clone() : 
				   new Point(base.width/2, base.height/2);
	    q = new JTextArea(3, cols);
	    q.setBorder(eb);
	    if ( s != null ) q.setText(s);
	    else q.setVisible(false);
	    q.setLineWrap(true);
	    q.setWrapStyleWord(true);
	    q.setOpaque(false);
	    q.setEditable(false);

	    setOpaque(false);
	    setBorder(eb);
	    add(q, BorderLayout.NORTH);
	    setPreferredSize(base);
	}

	/**
	 * Set the saying.
	 * @param s the new saying
	 */
	public void setText(String s) { 
	    if (s != null ) q.setText(s);
	    repaint();
	}

	/**
	 * Make the balloon visible or invisible.  Do not use isVisible,
	 * because overrirding it confuses java about redrawing the JPanel to
	 * hide the balloon.
	 * @param v make the balloon visible?
	 */
	public void setDisplayed(boolean v) { 
	    q.setVisible(v);
	    repaint();
	}

	/**
	 * Return the true if the balloon is displayed.  Don't use isVisible -
	 * see above.
	 * @return true if the balloon is displayed
	 */
	public boolean isDisplayed() { return q.isVisible(); }

	/**
	 * Paint the balloon onto the applet's Graphics object if it's visible.
	 * @param g the applet's Graphics object
	 */
	public void paintComponent(Graphics g) {
	    super.paintComponent(g);
	    if ( q.isVisible() ) {
		Rectangle qb = q.getBounds();
		int qx = qb.x + (3 * qb.width) / 4;
		int qy = qb.y + qb.height/2;
		Polygon p = new Polygon(new int[] { origin.x, qx, qx+10 }, 
					new int[] {origin.y, qy, qy },
					3);
		g.setColor(Color.white);
		g.fillPolygon(p);
		g.setColor(Color.black);
		g.drawPolygon(p);
		g.setColor(Color.white);
		g.fillRoundRect(qb.x, qb.y, qb.width, qb.height, 10, 10);
		g.setColor(Color.black);
		g.drawRoundRect(qb.x, qb.y, qb.width, qb.height, 10, 10);
	    }
	}

	/**
	 * Set the size of the word balloon.  Note that the new size becomes
	 * the preferred size.
	 * @param d the new size.
	 */
	public void setSize(Dimension d) {
	    setPreferredSize(d);
	    super.setSize(d);
	}
    }

    /**
     * Class to handle mouse clicks.  Put up the next clicked saying on a press
     * and take it down if putting it up succeeded.  Putting it up won't
     * succeed if another class of message is up on click.
     */
    class ClickHandler extends MouseAdapter {
	/** Setup was successful */
	boolean clearit;

	ClickHandler() { clearit = false; }

	public void mousePressed(MouseEvent e) {
	    clearit = showNext(CLICKED);
	}

	public void mouseReleased(MouseEvent e) {
	    if ( clearit ) hideSaying();
	}
    }

    /**
     * A class to return exponential interarrival times.
     */
    class Exponential {
	/** The underlying uniform random variable */
	Random r;
	/** The parameter of the exponential (1/mean) */
	double lambda;

	/**
	 * Create an Exponential rv with the given mean.
	 * @param mean the mean of the RV.
	 */
	Exponential(double mean) { 
	    r = new Random();
	    lambda = 1/mean;
	}

	/**
	 * Create an Exponential rv with the given mean.  Seed the random
	 * number generator.
	 * @param mean the mean of the RV.
	 * @param seed the seed for the random number generator.
	 */
	Exponential(double mean, long seed) {
	    r = new Random(seed);
	    lambda = 1/mean;
	}

	/**
	 * Return the next interarrival time.
	 */
	double nextTime() {
	    return -1.0 * Math.log(1.0 - r.nextDouble()) / lambda;
	}
    }

    /**
     * A TimerTask to display a new TIMED saying and schedule a HideTimed
     * tTimerTask holdTime milliseconds later.  If the initial showNext call
     * fails, a new ShowTimed TimerTask is scheduled instead.
     */
    class ShowTimed extends TimerTask {
	public void run() {
	    if ( showNext(TIMED) ) timer.schedule(new HideTimed(), holdTime);
	    else timer.schedule(this, (long) expTime.nextTime());
	}
    }

    /**
     * A TimerTask that hides the current saying and schedules a new timed
     * saying to be displayed.
     */
    class HideTimed extends TimerTask {
	public void run() {
	    hideSaying();
	    timer.schedule(new ShowTimed(), (long) expTime.nextTime());
	}
    }

    /** The timed saying info index */
    public static final int TIMED = 0;
    /** The clicked saying info index */
    public static final int CLICKED = 1;
    /** Number of lists to create */
    public static final int NUM_LISTS = 2;
    /** Manage this list randomly */
    public static final int RANDOM_LIST = 0;
    /** Manage this list serially */
    public static final int SERIAL_LIST = 1;

    /** The base image to display */
    Image image;
    /** An icon created from the scaled image that is actually painted.  */
    ImageIcon icon;
    /** The word balloon */
    WordBalloon w;
    /** Lists of potential timed sayings */
    Vector sayings[];
    /** next saying of each type to display */
    int nextSaying[];
    /** Way to move through each type of saying */
    int typeSaying[];
    /** Source of rendomness */
    Random rand;
    /** To schedule appearances and disappearances of the message */
    Timer timer;
    /** Give exponentially distributed times */
    Exponential expTime;
    /** time to hold up a timed saying */
    long holdTime;

    /**
     * Constructor - mostly straightforward initialization.  Note that if the
     * mean or hold time is less than 0, the first message in the first sayling
     * vector is statically displayed.
     * @param im the background image
     * @param s a array of saying Vectors.  The timed ones are in the Vector
     * s[0] and the ones cycles through by clicking are in s[1] (see
     * constants).
     * @param te array of types for the two lists indicating whether the
     * corresponding Vector of sayings should be traversed randomly or serially
     * (see constants)
     * @param o origin of the word balloon in panel coordinates
     * @param m mean of the exponential time distribution (in ms)
     * @param ht hold time (in ms)
     */
    public Mojo(Image im, Vector[] s, int[] ts, Point o, double m, long ht) { 
	super();
	image = im;
	setSize(getSize());	// This also scales the image and creates icon.
	sayings = new Vector[s.length];
	nextSaying = new int[s.length];
	typeSaying = new int[s.length];

	rand = new Random();

	// Copy the sayings and types in.
	for (int i = 0; i < s.length; i++) {
	    sayings[i] = (Vector) s[i].clone();
	    typeSaying[i] = (i < ts.length) ? ts[i] : RANDOM_LIST;
	    if ( sayings[i].size() > 0 && typeSaying[i] == RANDOM_LIST ) 
		nextSaying[i] = rand.nextInt(sayings[i].size());
	    else
		nextSaying[i] = 0;
	}
	holdTime = ht;

	// If either hold time or the mean is negative (or 0) the first message
	// in the first list should be displayed continuously, e.g., as an
	// error message.  Not allocating a timer keeps the message up and
	// initializing w to show the message puts it up.  If the message
	// shoudl be randomly played, no initial message is created and a timer
	// allocated.
	if ( m > 0.0 && holdTime > 0 ) {
	    timer = new Timer();
	    expTime = new Exponential(m);
	    w = new WordBalloon(null, o);
	    w.setDisplayed(false);
	}
	else {
	    timer = null;
	    expTime = null;
	    w = new WordBalloon((String) sayings[0].elementAt(0), o);
	    w.setDisplayed(false);
	}

	// Attach the balloon and a clicker
	add(w);
	addMouseListener(new ClickHandler());
    }

    /**
     * Custom painting - draw the icon on the applet's Graphics object. 
     * @param g the applet's Graphics object
     */
    public void paintComponent(Graphics g) {
	super.paintComponent(g);
	Dimension d = getSize();
	int h = icon.getIconHeight();
	int w = icon.getIconWidth();
	h = d.height/2 - h/2;
	w = d.width/2 - w/2;
	icon.paintIcon(this, g, h, w);
    }

    /**
     * Resize the Panel - this also resizes the image icon and word balloon.
     * @param d the new dimension
     */
    public void setSize(Dimension d) {
	super.setSize(d);
	Image ti = ( d.width != 0 || d.height!= 0 ) ?
	    image.getScaledInstance(d.width, d.height, Image.SCALE_FAST) :
	    image;

	icon = new ImageIcon(ti);
	if ( w != null ) w.setSize(d);
    }

    /**
     * Display a new saying of the given type on the word balloon.  This fails
     * and returns false if there is already a saying displayed in the balloon
     * or if there are no sayings of the given type.
     * @param n the type of saying to display (see constants)
     * @return true if the saying was displayed (and will be until
     * hideSaying()).
     */
    public boolean showNext(int n) {
	Vector v = sayings[n];
	final int type = typeSaying[n];
	final int size = v.size();

	if ( size == 0 ) return false;

	synchronized (w) {
	    if ( w.isDisplayed() ) return false;

	    String saying = (String) v.elementAt(nextSaying[n]);
		       
	    switch (type) {
		case RANDOM_LIST:
		    nextSaying[n] = rand.nextInt(size);
		    break;
		case SERIAL_LIST:
		    nextSaying[n] = (nextSaying[n] + 1 ) % size;
		    break;
	    }
	    w.setText(saying);
	    w.setDisplayed(true);
	}
	return true;
    }

    /**
     * Undisplay the word balloon, if a timer is running.  Otherwise the
     * message should stay up indefinitely.
     */
    public void hideSaying() { 
	if (timer != null ) {
	    synchronized (w) { 
		w.setDisplayed(false); 
	    } 
	}
    } 

    /**
     * Start the timer dispatching random saying changes if a timer has been
     * allocated.
     */
    public void startTimer() {
	if ( timer != null )
	    timer.schedule(new ShowTimed(), (long) expTime.nextTime());
    }

    /**
     * Stop a running timer, if allocated.
     */
    public void stopTimer() { if (timer != null ) timer.cancel(); }
}
