import java.awt.Dimension;
import java.awt.Image;
import java.awt.Point;

import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;

import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.IOException;

import java.util.Vector;

import javax.swing.JApplet;

import Mojo;

/**
 * This applet displays an image and periodically overlays a word balloon with
 * witty sayings.  The source of the balloon, the images, and the sayings are
 * all configurable.
 * @author Ted Faber &lt;faber@lunabase.org&gt;
 * @version 1.0
 */
public class MojoApplet extends JApplet {
    /** The panel to display the sayings */
    Mojo mojo;

    /**
     * Constructor - just set everything to some safe value
     */
    public MojoApplet() { mojo = null; }

    /**
     * Get the given type of sayings list from the given URL.  If the order
     * type is given intitalize that, too.  Finally initialize the first saying
     * for each list as well.  On failures to read any of the quotes, throw an
     * IOException or NullPointerException.  Note that a value of "none"
     * (case-insensitive) for quotes will leave the syaings list empty with no
     * error indication.
     * @param quotes the URL to read (if null defaults to "sayings")
     * @param order a string of either "serial" or "random" defaults to "random"
     * @param type the type of saying list to initialize (see constants)
     * @exception IOException error reading quotes
     */
    protected void getQuotes(String quotes, String order, int type, 
	Vector[] sayings, int[] typeSaying) 
	    throws IOException {
	String defaultURLPrefix = getCodeBase().toString();

	if ( quotes.equalsIgnoreCase("none") ) return;

	if ( quotes != null ) {
	    if ( quotes.indexOf(':') == -1 ) 
		quotes = defaultURLPrefix + quotes;
	}
	else quotes = defaultURLPrefix +"sayings";

	if ( order != null ) {
	    if ( order.equalsIgnoreCase("serial") ) 
		typeSaying[type] = Mojo.SERIAL_LIST; 
	    else 
		typeSaying[type] = Mojo.RANDOM_LIST;
	}
	else typeSaying[type] = Mojo.RANDOM_LIST;

	URL quoteURL = new URL(quotes);
	URLConnection c = quoteURL.openConnection();
	LineNumberReader in =
	    new LineNumberReader(new InputStreamReader(
					    c.getInputStream()));
	String s;

	while ( (s = in.readLine()) != null ) 
	    sayings[type].add(s);

    }

    /**
     * Read parameters to initialize various parts of the system.
     */
    public void init() {
	String defaultURLPrefix = getCodeBase().toString();
	String originX;
	String originY;
	String imageURL;
	String timeP;
	String errorMessage = null;
	Image bg = null;
	Point origin = null;
	Dimension d = getSize();
	Vector[] sayings = new Vector[Mojo.NUM_LISTS];
	int[] typeSaying = new int[Mojo.NUM_LISTS];
	double mean = -1.0;
	long holdTime = -1;

	for (int i =0; i < sayings.length; i++ ) 
	    sayings[i] = new Vector();
	

	// Unless both an x and y value for the origin of the word balloon
	// are specified, the origin will be the middle of the applet (and
	// the origin variable will stay null)
	if ( (originX = getParameter("OriginX")) != null && 
	     (originY = getParameter("OriginY")) != null ) {
	    Double dd = new Double(originX);

	    if ( dd.doubleValue() > 0.0 && dd.doubleValue() < 1.0) {
		// If the origin is a pair of doubles, the point is in
		// (0,1) coordiantes of the image (0,0) upper left -> (1,1)
		// lower right.
		origin = new Point(
			(int) (Double.parseDouble(originX) * d.width),
			(int) (Double.parseDouble(originY) * d.height)
		    );
	    }
	    else {
		// The origin is in absolute integer coordinates (with
		// respect to the applet)
		origin = new Point(Integer.parseInt(originX),
				   Integer.parseInt(originY));
	    }
	}

	// Get the image parameter and fetch the image from the given URL.
	// If loading fails, set the error message to say why.
	try {
	    if ( (imageURL = getParameter("image")) != null ) {
		if ( imageURL.indexOf(':') == -1 ) 
		    imageURL = defaultURLPrefix + imageURL;
	    }
	    else imageURL = defaultURLPrefix + "bg.gif";

	    bg = getImage(new URL(imageURL));
	}
	catch (NullPointerException e) {
	    errorMessage = "Bad Image (null pointer): " +e;
	}
	catch (MalformedURLException e) {
	    errorMessage = "Bad image url: " + e;
	}

	// Get the list of quotes from the URL given in the quotes
	// parameter.  If there's a problem, set errorMessage.  
	if ( errorMessage == null ) {
	    try {
		getQuotes(getParameter("timedsayings"), 
			getParameter("timedorder"), Mojo.TIMED, sayings,
			typeSaying);
		getQuotes(getParameter("clickedsayings"), 
			getParameter("clickedorder"), Mojo.CLICKED, sayings, 
			typeSaying);
	    }
	    catch (NullPointerException e) {
		errorMessage = "Can't get quotes (bad pointer): " +e;
	    }
	    catch (IOException e) {
		errorMessage = "Can't get quotes: " + e;
	    }
	}

	// Set timing parameters.  Note that hold and mean both default to
	// 10 secs.
	if (errorMessage == null ) {
	    if ( (timeP = getParameter("holdtime")) != null ) {
		try {
		    holdTime = (long) (Double.parseDouble(timeP) * 1000.0);
		}
		catch (NumberFormatException e ) {
		    holdTime = 10000;
		}
	    }

	    if ( (timeP = getParameter("meantime")) != null ) {
		try {
		    mean = (Double.parseDouble(timeP) * 1000.0);
		}
		catch (NumberFormatException e ) {
		    mean = 10000.0;
		}
	    }
	}
	if ( errorMessage != null ) {
	    // There's been a configuration error.  Set the mojo panel to
	    // display the error.  (mean and holdTime will still be at their
	    // initial negative values, which will cause the Mojo object to
	    // display the first message in the first vector statically).

	    sayings[0].add(0, errorMessage);
	}
	mojo = new Mojo(bg, sayings, typeSaying, origin, mean, holdTime);
	mojo.setSize(d);

	setContentPane(mojo);
    }

    /**
     * Start mojo displaying messages.
     */
    public void start() { mojo.startTimer(); }

    /**
     * Stop mojo from displaying messages.
     */
    public void stop() { mojo.hideSaying(); mojo.stopTimer(); }
}
